#********************************************************************************************************
# Name: Rally Racing Game Mode for Battlefield 2
# Author: Mach10
# Organization: Desert Conflict www.Desert-Conflict.org
# Version 0.2 
#----------------------------------------------------------------------------------------------------------------------------
# Conditions and Limitations: You may use this file locally, on listening and dedicated servers. You MAY NOT 
# copy, redistributed, post, modify this file with out the expressed permission from the Desert Conflict Modification
# Team. You may contact us and request these rights at: www.desert-conflict.org
#This software is provided "as is". In no event shall the author, authors or contributors be held
#liable for  any direct, indirect, incidental, special, exemplary, or consequential damages (including)
#but not limited to, procurement of substitute good or services; loss of use, data or profits; or 
#business interruption) however caused and on any theory of liability, whether in contract, strict liability
#arising in any way out of the use of this software, even if advised of the possibility of such damage.
#=====
# Purpose: The file creates a new game mode for Battlefield 2, Rally Racing.
#=====
# Game Mode Information: You must spawn on the Coalition team, and take a DPV from your non-cap to the other
# team's non-cap and back again. The player who completes 3 full laps will win.
#
#*********************************************************************************************************
#
# Edited by Elxx from the Battleracer v1.1 source code, with permission from the Desert Conflict team
# 
# New Game Mode Info:	You must spawn on the Coalition team, and drive your vehicle through all of the 
#						checkpoints and then back to the start/finish line, to denote one lap. The first 
#						player to complete the specified number of laps will win the race.
#
# Battleracer v1.22
# 5/24/06
#
#*********************************************************************************************************

# Set default values
shortRaceLaps = 4
longRaceLaps = 3
intRaceLaps = 2
raceCountdownTime = 100
statsCode = ""
exitPenalty = 15
exitThreshold = 4
enableFF = 1
handheldWeapons = 1
vehicleWeapons = 1

weaponsDisabled = 0

STATS_SERVER = "205.234.105.49"
STATS_PORT = 16570

BRVERSION = "1.22"

Top = 0
Middle = 1
Bottom = 2

displayWinnersTimerIndex = 0
numberOfPlayers = 0
raceStarted = False
startCountdown = 60
num_pts = 0
startTime = 0.0
mapSize = 16

raceIsEnded = 0

endRaceTimer = None
displayWinnersTimer = None
lastChanceTimer = None
penaltyTimer = None

player_timesplit = {}
player_totaltime = {}
player_bestlap = {}
player_timepenalty = {}
player_disqualified = {}
player_startcheck = {}
player_setExitPenalty = {}
player_setDQ = {}

player_currVehicle = {}
player_lastVehicle = {}

player_ids = {}

import host
import bf2
import math
import bf2.Timer
import sys
import socket
import md5
import inspect
import re
import ConfigParser
import errno
from game.scoringCommon import addScore, RPL
from bf2 import g_debug
from socket import *

g_controlPoints = [] # cache, as this map won't change

a_debug = 0
class writer:
	def __init__(self, filename ):
		self.stream = open( filename, 'a+')

	def write(self, str):
		self.stream.write(str)
		self.stream.flush()
if a_debug: sys.stdout = sys.stderr = writer( 'stats_debug.log' )
 
def ExceptionOutput():
	sys.stderr.write("\n" + "Exception Occured: " + str(sys.exc_info()[0]) + "\n")
	sys.stderr.write("Value: " + str(sys.exc_info()[1]) + "\n")	
	sys.stderr.write("Line:" + str(readline(inspect.getfile(sys.exc_info()[2]),
					sys.exc_info()[2].tb_lineno)) + "\n")	
	sys.stderr.write("Line #: " + str(sys.exc_info()[2].tb_lineno) + "\n")
	sys.stderr.write("File: " + str(inspect.getfile(sys.exc_info()[2])) + "\n" + "\n")

def readline(filename, lineno):
	filen = re.sub('\\\\', '/', filename)
	file = open(filen, 'rU')
	lines = file.readlines()
	file.close()
	linen = lineno - 1
	line = re.sub('\s+', ' ', lines[linen])
	return line

def init():
	# events hook
	host.registerGameStatusHandler(onGameStatusChanged)
	if host.sgl_getIsAIGame() == 1:
		host.sh_setEnableCommander(0)
	else:
		host.sh_setEnableCommander(0)
		
	if g_debug: print "gpm_rally.py initialized"
		
def deinit():
	bf2.triggerManager.destroyAllTriggers()
	global g_controlPoints
	g_controlPoints = []
	host.unregisterGameStatusHandler(onGameStatusChanged)
	if g_debug: print "gpm_rally.py uninitialized"

def onGameStatusChanged(status):

	global g_controlPoints
	global checkpointnum
	global num_pts
	global endOfRoundTimer
	global raceStarted
	global statsCode
	global raceCountdownTime
	global lastChanceTimer
	global penaltyTimer
	global enableFF
	global handheldWeapons
	global vehicleWeapons
	
	global mapSize
	global raceLaps

	num_pts = 0

	CleanUp(0)

	if status == bf2.GameStatus.Playing: 
		g_controlPoints = bf2.objectManager.getObjectsOfType('dice.hfe.world.ObjectTemplate.ControlPoint')
		for obj in g_controlPoints:
			radius = float(obj.getTemplateProperty('radius'))
			fteam = float(obj.cp_getParam('team'))
			if fteam == 2:
				checkpointnum = int(obj.getTemplateProperty('ControlPointID'))
				if checkpointnum == 1:
					id = bf2.triggerManager.createRadiusTrigger(obj, FinishTrigger, '<<PCO>>', radius, (1, 2, 3))
				elif checkpointnum == 2:
					id = bf2.triggerManager.createRadiusTrigger(obj, StartCheckTrigger, '<<PCO>>', radius, (1, 2, 3))
				else:
					id = bf2.triggerManager.createRadiusTrigger(obj, CheckpointTrigger, '<<PCO>>', radius, (1, 2, 3))
					num_pts += 1
				
				obj.triggerId = id
				obj.lastAttackingTeam = 0
		try:
			config = ConfigParser.ConfigParser()
			config.read("mods/battleracer2/settings/racesettings.ini")
			longRaceLaps = int(config.get("races", "longlaps"))
			shortRaceLaps = int(config.get("races", "shortlaps"))
			intRaceLaps = int(config.get("races", "intlaps"))
			raceCountdownTime = int(config.get("races", "racecountdowntime"))
			enableFF = int(config.get("weapons", "friendlyfire"))
			handheldWeapons = int(config.get("weapons", "handheldweapons"))
			vehicleWeapons = int(config.get("weapons", "vehicleweapons"))
		except:
			shortRaceLaps = 4
			longRaceLaps = 3
			intRaceLaps = 2
			raceCountdownTime = 100
			enableFF = 1
			handheldWeapons = 1
			vehicleWeapons = 1
			ExceptionOutput()
			
		try:
			config2 = ConfigParser.ConfigParser()
			config2.read("mods/battleracer2/settings/statsettings.ini")
			statsCode = str(config2.get("stats", "statscode"))
		except:
			statsCode = ""
			ExceptionOutput()
		
		if host.sgl_getMapName() == "seaside_raceway" or host.sgl_getMapName() == "seaside_night":
			if num_pts > 4:
				mapSize = 32
			else:
				mapSize = 16
		elif host.sgl_getMapName() == "mashtuur_city" or host.sgl_getMapName() == "strike_at_karkand":
			if num_pts == 7:
				mapSize = 64
			elif num_pts == 6:
				mapSize = 32
			else:
				mapSize = 16
		
		if mapSize == 64: 
			raceLaps = intRaceLaps
		if mapSize == 32:
			raceLaps = longRaceLaps
		if mapSize == 16:
			raceLaps = shortRaceLaps
		
		bf2.gameLogic.setTickets(1, 1)
		bf2.gameLogic.setTickets(2, 1)
		
		bf2.gameLogic.setTicketState(1, 0)
		bf2.gameLogic.setTicketState(2, 0)

		host.registerHandler('PlayerSpawn', onPlayerSpawn)
		host.registerHandler('EnterVehicle', onEnterVehicle)
		host.registerHandler('ExitVehicle', onExitVehicle)
		host.registerHandler('PlayerConnect', onPlayerConnect)
		host.registerHandler('PlayerDisconnect', onPlayerDisconnect)
		host.registerHandler('PlayerDeath', onPlayerDeath) 
		host.registerHandler('DropKit', onDropKit)
		
		raceStarted = False
		
		if not lastChanceTimer:
			lastChanceTimer = bf2.Timer(lastChanceRestart, 0 , 1)
			lastChanceTimer.setRecurring(10.0)
			
		if not penaltyTimer:
			penaltyTimer = bf2.Timer(checkPenalty, 0 , 1)
			penaltyTimer.setRecurring(1.0)
		
		# if there's not enough players, don't start the countdown yet
		if bf2.playerManager.getNumberOfPlayers() >= int(host.rcon_invoke('sv.numPlayersNeededToStart')):
			startTiming()
		
		generateIDs()
		
		if g_debug: print "Rally gamemode initialized."
		print 'Author is Mach10, mach10@comcast.net'
		print 'With additions by Elxx (http://elxx.net)'

	else:
		bf2.triggerManager.destroyAllTriggers()
		g_controlPoints = []
	
DOWNWARDS = -1
UPWARDS = 1

def startTiming():
	global startCountdown
	global raceStartTimer
	global raceCountdownTime
	
	disableWeapons()
	
	startCountdown = raceCountdownTime
	
	try: raceStartTimer
	except NameError:
		raceStartTimer = bf2.Timer(raceStarter, 0 , 1)
		raceStartTimer.setRecurring(1.0)

def formatTime(t):
	if t > 59:
		minutes = math.floor(t / 60)
		minutes_s = '%02d' % minutes
	else:
		minutes = 0
		minutes_s = '00'
	if t > 1:
		seconds = math.floor(t - (int(minutes) * 60))
		seconds_s = '%02d' % seconds
	else:
		seconds = 0
	mills = (t - (int(seconds) + (int(minutes) * 60))) * 100
	mills_s = '%02d' % mills
	return str(minutes_s) + ":" + str(seconds_s) + "." + str(mills_s)
	
def formatTimePart(t):
	if t > 59:
		minutes = math.floor(t / 60)
	else:
		minutes = 0
	if t > 1:
		seconds = math.floor(t - (int(minutes) * 60))
	else:
		seconds = 0
	mills = (t - (int(seconds) + (int(minutes) * 60))) * 100
	
	return int(minutes),int(seconds),int(mills)
	
def generateIDs():
	global player_ids
	for p in bf2.playerManager.getPlayers():
		player_ids[p.getName()] = p.index
		if not player_disqualified.has_key(p.getName()):
			player_disqualified[p.getName()] = 0
		if not player_startcheck.has_key(p.getName()):
			player_startcheck[p.getName()] = 0
		if not player_currVehicle.has_key(p.getName()):
			player_currVehicle[p.getName()] = 0
		if not player_lastVehicle.has_key(p.getName()):
			player_lastVehicle[p.getName()] = 0
		if not player_setExitPenalty.has_key(p.getName()):
			player_setExitPenalty[p.getName()] = 0

_swap2 = lambda (x,y): (y,x)

def sortbyvalue(dict):
	mdict = map(_swap2, dict.items())
	mdict.sort()
	mdict = map(_swap2, mdict)
	return mdict

def raceStarter(data):
	global startCountdown
	global raceStarted
	global startTime
	global player_timesplit
	global player_startcheck
		
	startCountdown -= 1
	
	if startCountdown > 0:
		bf2.gameLogic.setTickets(1, startCountdown)
		bf2.gameLogic.setTickets(2, startCountdown)
	else:
		bf2.gameLogic.setTickets(1, raceLaps)
		bf2.gameLogic.setTickets(2, raceLaps)
	
	if startCountdown == 110:
		sendMsg("The race will start in 2 minutes. Drive to the start line and prepare your vehicle.")
	if startCountdown == 90:
		sendMsg("The race will start in about 90 seconds. Drive to the start line and prepare your vehicle.")
	if startCountdown == 60:
		sendMsg("The race will start in 1 minute. Drive to the start line and prepare your vehicle.")
	if startCountdown == 30:
		sendMsg("The race will begin in 30 seconds.")
	if startCountdown == 10:
		sendMsg("The race will begin in 10 seconds. Gentlemen, start your engines!")
		sendMsg("This will be a " + str(raceLaps) + "-lap race.")
		if enableFF:
			if handheldWeapons:
				sendMsg("Handheld weapons are enabled.")
			else:
				sendMsg("Handheld weapons are DISABLED.")
			if vehicleWeapons:
				sendMsg("Vehicle-mounted weapons are enabled.")
			else:
				sendMsg("Vehicle-mounted weapons are DISABLED.")
		else:
			sendMsg("Weapons and Friendly Fire are disabled for this race. Please drive cleanly.")
	if startCountdown == 3:
		sendEmphMsg("   3            3            3            3")
	if startCountdown == 2:
		sendEmphMsg("   2            2            2            2")
	if startCountdown == 1:
		sendEmphMsg("   1             1             1             1")
	if startCountdown == 0:
		sendEmphMsg("   GO          GO           GO           GO")
		raceStarted = True
		startTime = host.timer_getWallTime()
		controlPoints = bf2.objectManager.getObjectsOfType('dice.hfe.world.ObjectTemplate.ControlPoint')
		try:
			for cp in controlPoints:
				ptid = int(cp.getTemplateProperty('ControlPointID'))
				if ptid == 2:
					pcos = bf2.triggerManager.getObjects(cp.triggerId)
					for o in pcos:
						if not o: continue 
						if o.getParent(): continue
						occupyingPlayers = o.getOccupyingPlayers()
						for p in occupyingPlayers:
							if p != occupyingPlayers[0]: continue
							if p.isAlive() and not p.isManDown():
								if not p.killed:
									if p.getDefaultVehicle() != p.getVehicle():
										if player_disqualified[p.getName()] < 3:
											player_startcheck[p.getName()] = 1
		except:
			ExceptionOutput()
	if startCountdown <= -5:
		enableWeapons()

def sendMsg(msg):
	host.rcon_invoke("game.sayAll \"" + str(msg) + "\"")

def enableWeapons():
	global weaponsDisabled
	
	disableWeapons()
	if vehicleWeapons:
		host.rcon_invoke("ObjectTemplate.active HMG_M134_Gun")
		host.rcon_invoke("ObjectTemplate.projectileTemplate HMG_M134_Projectile")
		host.rcon_invoke("ObjectTemplate.velocity 500")
		host.rcon_invoke("ObjectTemplate.active HMG_M134_Projectile")
		host.rcon_invoke("ObjectTemplate.detonation.explosionDamage 200")
		host.rcon_invoke("ObjectTemplate.active HMG_M2HB")
		host.rcon_invoke("ObjectTemplate.projectileTemplate HMG_M2HB50cal_Projectile")
		host.rcon_invoke("ObjectTemplate.velocity 500")
	if handheldWeapons:
		host.rcon_invoke("ObjectTemplate.active kni_knife")
		host.rcon_invoke("ObjectTemplate.projectileTemplate kni_knife_Projectile")
		host.rcon_invoke("ObjectTemplate.velocity 10")
		host.rcon_invoke("ObjectTemplate.active uspis_92fs")
		host.rcon_invoke("ObjectTemplate.projectileTemplate USPIS_92FS_Projectile")
		host.rcon_invoke("ObjectTemplate.velocity 1000")
	if enableFF:
		host.rcon_invoke("sv.soldierFriendlyFire 100")
		host.rcon_invoke("sv.vehicleFriendlyFire 100")
	else:
		host.rcon_invoke("sv.soldierFriendlyFire 0")
		host.rcon_invoke("sv.vehicleFriendlyFire 0")
	host.rcon_invoke("sv.soldierSplashFriendlyFire 100")
	host.rcon_invoke("sv.vehicleSplashFriendlyFire 100")
	weaponsDisabled = 0
	
def disableWeapons():
	global weaponsDisabled
	
	host.rcon_invoke("ObjectTemplate.active HMG_M134_Gun")
	host.rcon_invoke("ObjectTemplate.projectileTemplate HMG_M134_Projectile")
	host.rcon_invoke("ObjectTemplate.velocity 1")
	host.rcon_invoke("ObjectTemplate.active HMG_M134_Projectile")
	host.rcon_invoke("ObjectTemplate.detonation.explosionDamage 0")
	host.rcon_invoke("ObjectTemplate.active HMG_M2HB")
	host.rcon_invoke("ObjectTemplate.projectileTemplate HMG_M2HB50cal_Projectile")
	host.rcon_invoke("ObjectTemplate.velocity 0")
	host.rcon_invoke("ObjectTemplate.active kni_knife")
	host.rcon_invoke("ObjectTemplate.projectileTemplate kni_knife_Projectile")
	host.rcon_invoke("ObjectTemplate.velocity 0")
	host.rcon_invoke("ObjectTemplate.active uspis_92fs")
	host.rcon_invoke("ObjectTemplate.projectileTemplate USPIS_92FS_Projectile")
	host.rcon_invoke("ObjectTemplate.velocity 1")
	host.rcon_invoke("sv.soldierFriendlyFire 0")
	host.rcon_invoke("sv.vehicleFriendlyFire 0")
	host.rcon_invoke("sv.soldierSplashFriendlyFire 100")
	host.rcon_invoke("sv.vehicleSplashFriendlyFire 100")
	host.rcon_invoke("sv.spawnTime 4")
	weaponsDisabled = 1
	
def sendEmphMsg(msg):
	sendMsg("================================")
	sendMsg("================================")
	sendMsg(msg)
	sendMsg("================================")
	sendMsg("================================")

def StartCheckTrigger(triggerId, cp, vehicle, enter, userData):	
	global raceStarted
	global startTime
	global player_disqualified
	global player_startcheck

	if raceStarted:
		pcos = bf2.triggerManager.getObjects(cp.triggerId)
		ptid = int(cp.getTemplateProperty('ControlPointID'))
		for o in pcos:
			if not o: continue 
			if o.getParent(): continue
			occupyingPlayers = o.getOccupyingPlayers()
			for p in occupyingPlayers:
				if p != occupyingPlayers[0]: continue
				if p.isAlive() and not p.isManDown():
					if not p.killed:
						if p.getDefaultVehicle() != p.getVehicle():
							if player_disqualified[p.getName()] < 3:
								player_startcheck[p.getName()] = 1

def CheckpointTrigger(triggerId, cp, vehicle, enter, userData):	
	global raceStarted
	global startTime
	global player_disqualified

	if raceStarted:
		pcos = bf2.triggerManager.getObjects(cp.triggerId)
		ptid = int(cp.getTemplateProperty('ControlPointID'))
		for o in pcos:
			if not o: continue 
			if o.getParent(): continue
			occupyingPlayers = o.getOccupyingPlayers()
			for p in occupyingPlayers:
				if p != occupyingPlayers[0]: continue
				if p.isAlive() and not p.isManDown():
					if not p.killed:
						if p.getDefaultVehicle() != p.getVehicle():
							if (p.score.cpCaptures + 1) < (ptid - 2):
								playerDriveReverse(p, vehicle)
							#if (p.score.cpCaptures) == (ptid - 2):
							#	PlayerDriveReverse(p, vehicle)
							if player_disqualified[p.getName()] < 3:
								if p.getTeam() == 2 and (p.score.cpCaptures) == (ptid - 3) and p.score.kills < raceLaps and p.score.revives > 0:
									passTime = 0
									if player_timesplit.has_key(p.getName()):
										passTime = host.timer_getWallTime() - player_timesplit[p.getName()]
									if not player_timesplit.has_key(p.getName()):
										passTime = host.timer_getWallTime() - startTime
									sendMsg(p.getName() + " has passed Checkpoint " + str(ptid - 2) + " * split " + str(formatTime(passTime)))
									p.score.cpCaptures += 1
									addScore(p, 1, RPL)


def FinishTrigger(triggerId, cp, vehicle, enter, userData):				
	global displayWinnersTimerIndex
	global num_pts
	global raceLaps
	global raceStarted
	global player_bestlap
	global player_totaltime
	global player_timesplit
	global player_timepenalty
	global player_disqualified
	global player_startcheck

	pcos = bf2.triggerManager.getObjects(cp.triggerId)
	for o in pcos:
		if not o: continue
		if o.getParent(): continue
		occupyingPlayers = o.getOccupyingPlayers()
		for p in occupyingPlayers:
			if p != occupyingPlayers[0]: 
				continue
			if p.isAlive() and not p.isManDown():
				if not p.killed:
					if p.getDefaultVehicle() != p.getVehicle():
						if raceStarted:
							if not player_startcheck[p.getName()]:
								sendMsg("FALSE START - " + str(p.getName()))
								if not player_timepenalty.has_key(p.getName()):
									player_timepenalty[p.getName()] = 0.0
								player_timepenalty[p.getName()] += 20.0
							if player_disqualified[p.getName()] < 3:
								# Lap completed
								if p.getTeam() == 2 and (p.score.cpCaptures) == (num_pts) and p.score.cpCaptures != 0:
									p.score.cpCaptures = 0
									p.score.kills += 1
									addScore(p, 1, RPL)
									
									totalTime = host.timer_getWallTime() - startTime
									passTime = 0

									if player_timesplit.has_key(p.getName()):
										passTime = host.timer_getWallTime() - player_timesplit[p.getName()]
									if not player_timesplit.has_key(p.getName()):
										passTime = host.timer_getWallTime() - startTime

									player_timesplit[p.getName()] = host.timer_getWallTime()

									if player_bestlap.has_key(p.getName()):
										if passTime < player_bestlap[p.getName()]:
											player_bestlap[p.getName()] = passTime
									else:
										player_bestlap[p.getName()] = passTime

									sendMsg(p.getName() + " has completed lap " + str(p.score.kills) + "/" + str(raceLaps) + " * split " + str(formatTime(passTime)) + " * total " + str(formatTime(totalTime)))

									if p.score.kills == raceLaps:
										if player_timepenalty.has_key(p.getName()):
											totalTime += player_timepenalty[p.getName()]
										player_totaltime[p.getName()] = totalTime
										try: bf2.gameLogic.sendGameEvent(p, 10, 1) # "You have finished the race." --> HUD_HUD_REPLENISH_GOT_REPAIR_POINT
										except:
											ExceptionOutput()
										checkWinners(p)
								if (p.score.cpCaptures == 0) and (not player_totaltime.has_key(p.getName())):
									p.score.revives = 1
									addScore(p, 1, RPL)
						else: 
							#The race hasn't started yet, player is going beyond the start line, so warn him
							try: bf2.gameLogic.sendGameEvent(p, 10, 0) # "Get behind the start line or you will be disqualified!" --> HUD_HUD_REPLENISH_GOT_HEALING_POINT
							except:
								ExceptionOutput()

def checkWinners(player):
	global endRaceTimer
	global player_totaltime
	global displayWinnersTimer
	
	disableWeapons()
	
	numberOfPlayers = host.pmgr_getNumberOfPlayers()
	
	if len(player_totaltime) >= numberOfPlayers:
		if not displayWinnersTimer:
			displayWinnersTimer = bf2.Timer(displayWinners, 0, 1)
			displayWinnersTimer.setRecurring(3.0)
	else:
		if not endRaceTimer:
			endRaceTimer = bf2.Timer(endRaceTime, 0, 1)
			endRaceTimer.setRecurring(60.0)

def endRaceTime(data):
	global endRaceTimer
	global displayWinnersTimer
	global raceIsEnded
	
	if raceIsEnded > 0:
		if not displayWinnersTimer:
			displayWinnersTimer = bf2.Timer(displayWinners, 0, 1)
			displayWinnersTimer.setRecurring(3.0)
		endRaceTimer.destroy()
		endRaceTimer = None
	else:
		raceIsEnded = 1

def displayWinners(data):
	global displayWinnersTimerIndex
	global player_totaltime
	global player_bestlap
	global displayWinnersTimer
	global player_ids
	global statsCode
	
	displayWinnersTimerIndex += 1
	
	# Driver name is stored as x[0], total race time as x[1]
	sortedplayer_totaltime = sortbyvalue(player_totaltime)
	
	if displayWinnersTimerIndex < 3:
		timesIndexed = 0
		sendMsg("Top 3:")
		for x in sortedplayer_totaltime:
			timesIndexed = timesIndexed + 1
			if timesIndexed < 4:
				if timesIndexed > 1:
					try: fpDiff = " (+" + str(formatTime(x[1] - sortedplayer_totaltime[0][1])) + ")"
					except:
						ExceptionOutput()
				else:
					fpDiff = ""
				try: sendMsg(str(timesIndexed) + ". " + str(x[0].replace("<","").replace(">","")) + " * " + str(formatTime(x[1])) + fpDiff)
				except:
					ExceptionOutput()
		sendMsg("==================================")
		if timesIndexed < 3:
			sendMsg("==================================")
		if timesIndexed < 2:
			sendMsg("==================================")
	
	elif displayWinnersTimerIndex < 7:
		timesIndexed = 0
		sendMsg("Race Results:")
		for x in sortedplayer_totaltime:
			timesIndexed = timesIndexed + 1
			if timesIndexed > 1:
				try: fpDiff = " (+" + str(formatTime(x[1] - sortedplayer_totaltime[0][1])) + ")"
				except:
					ExceptionOutput()
			else:
				fpDiff = ""
			try: sendMsg(str(timesIndexed) + ". " + str(x[0].replace("<","").replace(">","")) + " * " + str(formatTime(x[1])) + fpDiff)
			except:
				ExceptionOutput()
		if timesIndexed < 4:
			sendMsg("==================================")
		if timesIndexed < 3:
			sendMsg("==================================")
		if timesIndexed < 2:
			sendMsg("==================================")
		
	elif displayWinnersTimerIndex < 10:
		sortedplayer_bestlap = sortbyvalue(player_bestlap)
		timesIndexed = 0
		sendMsg("Best Lap Times:")
		# Driver name is stored as x[0], best time as x[1]
		for x in sortedplayer_bestlap:
			timesIndexed = timesIndexed + 1
			if timesIndexed < 4:
				try: sendMsg(str(formatTime(x[1])) + " - " + str(x[0].replace("<","").replace(">","")))
				except:
					ExceptionOutput()
		if timesIndexed < 3:
			sendMsg("==================================")
		if timesIndexed < 2:
			sendMsg("==================================")
		sendMsg("==================================")
		
	elif displayWinnersTimerIndex == 10:
		if statsCode:
			sendStats(STATS_SERVER, STATS_PORT)
		timesIndexed = 0
		for x in sortedplayer_totaltime:
			try: x = bf2.PlayerManager.Player(player_ids[x[0]])
			except:
				ExceptionOutput()
			else:
				timesIndexed = timesIndexed + 1
				if timesIndexed == 1:
					try: bf2.gameLogic.sendMedalEvent(x, 1031406, 1)
					except:
						ExceptionOutput()
				if timesIndexed == 2:
					try: bf2.gameLogic.sendMedalEvent(x, 1031406, 2)
					except:
						ExceptionOutput()
				if timesIndexed == 3:
					try: bf2.gameLogic.sendMedalEvent(x, 1031406, 3)
					except:
						ExceptionOutput()
		
	elif displayWinnersTimerIndex == 13:
		sendMsg("Ending race...")
		
	elif displayWinnersTimerIndex == 14:
		for p in bf2.playerManager.getPlayers():
			try: p.score.score = 0
			except:
				ExceptionOutput()
		timesIndexed = 0
		p_score = {}
		for x in sortedplayer_totaltime:
			timesIndexed = timesIndexed + 1
			if x[1] != "DNF":
				if timesIndexed == 1:
					p_score[x[0]] = 10
				if timesIndexed == 2:
					p_score[x[0]] = 8
				if timesIndexed == 3:
					p_score[x[0]] = 6
				if timesIndexed >= 4 and timesIndexed <= 8:
					p_score[x[0]] = 9 - timesIndexed
				if timesIndexed > 8:
					p_score[x[0]] = 0
			else:
				p_score[x[0]] = 0
		for p in bf2.playerManager.getPlayers():
			if player_totaltime.has_key(p.getName()):
				p.score.score = p_score[p.getName()]
				if player_totaltime[p.getName()] != "DNF":
					try:
						m,s,l = formatTimePart(player_totaltime[p.getName()])
						p.score.cpCaptures = 0
						p.score.rplScore = m
						p.score.skillScore = 0
						p.score.kills = s
						p.score.deaths = l
					except:
						ExceptionOutput()
				else:
					try:
						p.score.score = 0
						p.score.cpCaptures = 0
						p.score.rplScore = 0
						p.score.skillScore = 0
						p.score.kills = 0
						p.score.deaths = 0
					except:
						ExceptionOutput()
			else:
				try:
					p.score.score = 0
					p.score.cpCaptures = 0
					p.score.rplScore = 0
					p.score.skillScore = 0
					p.score.kills = 0
					p.score.deaths = 0
				except:
					ExceptionOutput()
		CleanUp(1)
		
	elif displayWinnersTimerIndex > 15:
		CleanUp(1)
	
def lastChanceRestart(data):
	global displayWinnersTimerIndex
	if displayWinnersTimerIndex > 15 and raceStarted:
		CleanUp(1)
	if raceStarted and not weaponsDisabled:
		if enableFF:
			host.rcon_invoke("sv.soldierFriendlyFire 100")
			host.rcon_invoke("sv.vehicleFriendlyFire 100")
		else:
			host.rcon_invoke("sv.soldierFriendlyFire 0")
			host.rcon_invoke("sv.vehicleFriendlyFire 0")
	host.rcon_invoke("sv.soldierSplashFriendlyFire 100")
	host.rcon_invoke("sv.vehicleSplashFriendlyFire 100")
	
def CleanUp(end_race):
	global displayWinnersTimerIndex
	global numberOfPlayers
	global numberOfPlayersChecked 
	global player_timesplit
	global player_totaltime
	global player_bestlap
	global player_timepenalty
	global player_disqualified
	global endRaceTimer
	global displayWinnersTimer
	global raceIsEnded
	global player_startcheck
	global player_setExitPenalty
	global player_setDQ
	global player_currVehicle
	global player_lastVehicle
	player_startcheck = None
	player_startcheck = {}
	player_setExitPenalty = None
	player_setExitPenalty = {}
	player_setDQ = None
	player_setDQ = {}
	player_currVehicle = None
	player_currVehicle = {}
	player_lastVehicle = None
	player_lastVehicle = {}
	if endRaceTimer:
		try: endRaceTimer.destroy()
		except:
			ExceptionOutput()
	if displayWinnersTimer:
		try: displayWinnersTimer.destroy()
		except:
			ExceptionOutput()
	endRaceTimer = None
	displayWinnersTimer = None
	displayWinnersTimerIndex = 0
	numberOfPlayers = 0
	numberOfPlayersChecked = 0
	raceIsEnded = 0
	player_timesplit = None
	player_totaltime = None
	player_bestlap = None
	player_timepenalty = None
	player_disqualified = None
	player_timesplit = {}
	player_totaltime = {}
	player_bestlap = {}
	player_timepenalty = {}
	player_disqualified = {}
	if end_race == 1:
		EndRace()

def EndRace():
	host.sgl_endGame(2, 3)
	
def sendStats(sServer, sPort):
	global player_totaltime
	global player_bestlap
	global player_ids
	global mapSize
	global raceLaps
	global statsCode
	
	message = ""
	
	print "Creating race data"
	message += statsCode + "%bR%" + str(BRVERSION) + "%bR%" + host.sgl_getMapName() + "%bR%" + str(mapSize) + "%bR%" + str(raceLaps) + "%bR%" + str(enableFF) + "%bR%" + str(handheldWeapons) + "%bR%" + str(vehicleWeapons) + "%Rb%"
	
	for p in bf2.playerManager.getPlayers():
		try: player_totaltime[p.getName()]
		except:
			player_totaltime[p.getName()] = "DNF"
		try: player_bestlap[p.getName()]
		except:
			player_bestlap[p.getName()] = "DNF"
	
	print "Sorting times"
	sortedplayer_totaltime = sortbyvalue(player_totaltime)
	
	for p in sortedplayer_totaltime:
		m = md5.new()
		m.update(str(p[0]) + str(p[1]) + "feads5")
		try: x = bf2.PlayerManager.Player(player_ids[p[0]])
		except:
			ExceptionOutput()
			phash = "na"
		else:
			phash = x.getProfileId()
		# Hex - Keyhash - Name - TotalTime - BestLap
		try: 
			message += m.hexdigest() + "%bR%" + str(phash) + "%bR%" + str(p[0]) + "%bR%" + str(p[1]) + "%bR%" + str(player_bestlap[p[0]]) + "%sP%"
		except: 
			ExceptionOutput()
	
	print "Creating socket"
	sock = socket(AF_INET, SOCK_STREAM)
	print "Setting to non-blocking"
	sock.settimeout(1.0)
	print "Connecting to stat server"
	try:
		sock.connect((sServer, sPort))
	except:
		sendMsg("Could not connect to stat server")
		ExceptionOutput()
		sock.close()
		return 0
	print "Sending message"
	messlen = 0
	tries = 0
	fail = 0
	data = ""
	while messlen < len(message):
		tries += 1
		if tries >= 2:
			print "Failed to send complete message"
			fail = 1
			break
		try: messlen += sock.send(message)
		except:
			ExceptionOutput()
		if messlen != len(message):
			print "Send looping..."
	
	if fail == 0:
		sendMsg("Race data sent to global stat server - http://br.elxx.net/")
		'''print "Getting response"
		data = ""
		try: data = sock.recv(1024)
		except:
			data = "error"
			ExceptionOutput()
		print "Checking result"
		if data != "error":
			try:
				exec(data)
			else:
				ExceptionOutput()'''
	
	print "Closing socket"
	sock.close()
	
def onPlayerConnect(p):
	print "Player connected"
	generateIDs()
	# dont teamswitch alive players, or they will have the wrong teams kit
	if p.isAlive(): return
	p.setTeam(2)
	
def onPlayerDisconnect(p):
	print "Player disconnected"
	generateIDs()
	
def onPlayerDeath(p, soldierObject):
	global raceStarted
	global player_timepenalty
	global player_lastVehicle
	if raceStarted:
		player_lastVehicle[p.getName()] = 0
		if p.getTeam() == 2:
			p.score.cpCaptures = 0
			addScore(p, 1, RPL)
			if p.getSuicide():
				if p.getTimeToSpawn() != 0:
					p.setTimeToSpawn(8)
	
def displayPlayerPOS(data):
	for p in bf2.playerManager.getPlayers():
		if (p.getName() == "gunzip") and p.isAlive():
			vehicle = p.getVehicle()
			pPos = vehicle.getPosition()
			print pPos

def onPlayerSpawn(player, soldier):
	print "Player spawned"
	global player_disqualified
	player.killed = False
	if player_disqualified[player.getName()] >= 2:
		if player.getTeam() == 2:
			try: bf2.gameLogic.sendGameEvent(player, 10, 3) # "You must spawn as a Spectator for the duration of this race." --> HUD_HUD_REPLENISH_DRIVER
			except:
				ExceptionOutput()
			if hasattr(player.getDefaultVehicle(), 'setDamage'):
				try: player.getDefaultVehicle().setDamage(-1)
				except:
					ExceptionOutput()
				else: player.setTeam(1)

def onEnterVehicle(player, vehicle, freeSoldier = False):
	global player_currVehicle
	updateCaptureStatus(vehicle)
	if bf2.objectManager.getRootParent(vehicle).templateName == "jeep_faav" or bf2.objectManager.getRootParent(vehicle).templateName == "faav_night":
		if player.getTeam() == 2:
			if raceStarted:
				if player_lastVehicle[player.getName()]:
					if bf2.objectManager.getRootParent(vehicle) != bf2.objectManager.getRootParent(player_lastVehicle[player.getName()]):
						player_setExitPenalty[player.getName()] = 1
			player_currVehicle[player.getName()] = vehicle
			player_lastVehicle[player.getName()] = 0
		else:
			if hasattr(player.getDefaultVehicle(), 'setDamage'):
				try: player.getDefaultVehicle().setDamage(-1)
				except:
					ExceptionOutput()
			if hasattr(vehicle, 'setDamage'):
				try: vehicle.setDamage(1)
				except:
					ExceptionOutput()

def onExitVehicle(player, vehicle):
	global raceStarted
	global player_totaltime
	global player_timepenalty
	global player_disqualified
	global player_currVehicle
	print "Player exiting ", player.getName()
	player_currVehicle[player.getName()] = 0
	player_lastVehicle[player.getName()] = vehicle
	cp = getOccupyingCP(player)
	player.setIsInsideCP(cp != None and cp.cp_getParam('unableToChangeTeam') == 0)
	updateCaptureStatus(vehicle)
	
def playerDriveReverse(player, vehicle):
	global player_disqualified
	if not player_totaltime.has_key(p.getName()):
		player_disqualified[player.getName()] = 2
		sendMsg(str(player.getName()) + "disqualified for driving around track backwards.")
		if hasattr(vehicle, 'setDamage'):
			try: 
				vehicle.setDamage(1)
			except:
				ExceptionOutput()
		if hasattr(player.getDefaultVehicle(), 'setDamage'):
			try: 
				player.getDefaultVehicle().setDamage(-1)
			except:
				ExceptionOutput()

def onDropKit(p, k):
	if p.getTeam() == 1:
		if p.isAlive():
			if hasattr(p.getDefaultVehicle(), 'setDamage'):
				try: 
					p.getDefaultVehicle().setDamage(-1)
				except:
					ExceptionOutput()

def checkPenalty(data):
	dq_checker = 0
	for player in bf2.playerManager.getPlayers():
		if player_disqualified.has_key(player.getName()):
			if player_disqualified[player.getName()] < 2:
				dq_checker = 1
		if player_setExitPenalty[player.getName()]:
			print "Penalizing player"
			if player_disqualified[player.getName()] < 2:
				sendMsg("15-second penalty assessed to " + str(player.getName()) + " for changing vehicles")
				if not player_timepenalty.has_key(player.getName()):
					player_timepenalty[player.getName()] = 0.0
				player_timepenalty[player.getName()] += 15.0
				player_disqualified[player.getName()] += 1
			if player_disqualified[player.getName()] == 2:
				print "DQing player"
				sendMsg(str(player.getName()) + " has been disqualified from this race")
				try: bf2.gameLogic.sendGameEvent(player, 10, 2) # "You have been disqualified from the race." --> HUD_HUD_REPLENISH_GOT_AMMO_POINT
				except:
					ExceptionOutput()
				if player.getVehicle() == player.getDefaultVehicle():
					print "Soldier DQ"
					if hasattr(player.getDefaultVehicle(), 'setDamage'):
						try: 
							player.getDefaultVehicle().setDamage(-1)
						except:
							ExceptionOutput()
				else:
					occupyingPlayers = player.getVehicle().getOccupyingPlayers()
					if player == occupyingPlayers[0]: 
						print "Vehicle DQ"
						if hasattr(player.getVehicle(), 'setDamage'):
							try: 
								player.getVehicle().setDamage(1)
							except:
								ExceptionOutput()
			player_setExitPenalty[player.getName()] = 0
	if dq_checker == 0 and bf2.playerManager.getNumberOfPlayers() >= 1:
		CleanUp(1)
		
def onPlayerKilled(victim, attacker, weapon, assists, object):
	# check if killed by vehicle in motion
	if weapon == None and object != None:
		if hasattr(object, 'lastDrivingPlayerIndex'):
			attacker = bf2.playerManager.getPlayerByIndex(object.lastDrivingPlayerIndex)
			killedByEmptyVehicle = True
	if attacker == victim:
		victim.setTimeToSpawn(5)
		if killedByEmptyVehicle and object.getIsWreck():
			return
		victim.score.cpCaptures = 0
	elif attacker.getTeam() == 2 and victim.getTeam() == 2:
		if victim.getTimeToSpawn() != 0:
			victim.setTimeToSpawn(5)
	else:
		victim.setTimeToSpawn(5)

def updateCaptureStatus(vehicle):	
	rootVehicle = bf2.objectManager.getRootParent(vehicle)
	playersInVehicle = rootVehicle.getOccupyingPlayers()
	for p in playersInVehicle:
		print "Players in vehicle ", p.getName()
		cp = getOccupyingCP(p)
		p.setIsInsideCP(cp != None and cp.cp_getParam('unableToChangeTeam') == 0 and p == playersInVehicle[0])
	
def getOccupyingCP(player):
	vehicle = player.getVehicle()
	playerPos = vehicle.getPosition()
	
	closestCP = None
	if len(g_controlPoints) == 0: return None
	for obj in g_controlPoints:
		distanceTo = getVectorDistance(playerPos, obj.getPosition())
		if closestCP == None or distanceTo < closestCPdist:
			closestCP = obj
			closestCPdist = distanceTo
	
	pcos = bf2.triggerManager.getObjects(closestCP.triggerId)
	for o in pcos:
		if o == player.getDefaultVehicle():
			return closestCP
		else:
			for p in o.getOccupyingPlayers():
				if p == player:
					return closestCP
	
	return None

def getVectorDistance(pos1, pos2):
	diffVec = [0.0, 0.0, 0.0]
	diffVec[0] = math.fabs(pos1[0] - pos2[0])
	diffVec[1] = math.fabs(pos1[1] - pos2[1])
	diffVec[2] = math.fabs(pos1[2] - pos2[2])
	return math.sqrt(diffVec[0] * diffVec[0] + diffVec[1] * diffVec[1] + diffVec[2] * diffVec[2])